# This file is adapted from python code released by WellDone International
# under the terms of the LGPLv3.  WellDone International's contact information is
# info@welldone.org
# http://welldone.org
#
# Modifications to this file from the original created at WellDone International 
# are copyright Arch Systems Inc.

from command import Command
from iotile.core.hw.exceptions import *
import base64

class RPCCommand (Command):
    def __init__(self, address, feature, command, *args):
        """
        Create an RPC command 
        """

        self.addr = int(address)
        self.feat = int(feature)
        self.cmd = int(command)
        self.spec = 0

        self.args = args
        self.result = ""

    def __str__(self):
        args = self._format_args()
        header = bytearray(4)
        header[0] = self.addr
        header[1] = self.feat
        header[2] = self.cmd
        header[3] = self.spec

        packet = header + args

        cmd = "binrpc %s" % base64.standard_b64encode(packet)
        return cmd

    def _convert_int(self, arg):
        out = bytearray(2)

        out[0] = arg & 0xFF
        out[1] = (arg & 0xFF00) >> 8;

        converted = out[0] | (out[1] << 8)

        if converted != arg:
            raise ValueError("Integer argument was too large to fit in an rpc 16 bit int: %d" % arg)

        return out

    def _pack_arg(self, arg):
        if isinstance(arg, int) or isinstance(arg, long):
            return self._convert_int(arg), False
        elif isinstance(arg, bytearray):
            return arg, True
        elif isinstance(arg, basestring):
            return bytearray(arg), True

        raise ValueError("Unknown argument type could not be converted for rpc call.")

    def _format_args(self):
        fmtd = bytearray()

        num_ints = 0
        num_bufs = 0

        for arg in self.args:
            a,is_buf = self._pack_arg(arg)
            fmtd += a

            if is_buf:
                num_bufs += 1
                buff_len = len(a)

            if not is_buf:
                if num_bufs != 0:
                    raise ValueError("Invalid rpc parameters, integer came after buffer.")

                num_ints += 1

        if num_bufs > 1:
            raise ValueError("You must pass at most 1 buffer. num_bufs=%d" % num_bufs)

        if len(fmtd) > 20:
            raise ValueError("Arguments are greater then the maximum mib packet size, size was %d" % len(fmtd))

        #Calculate the command type spec
        self.spec = len(fmtd)

        if len(fmtd) < 20:
            fmtd += bytearray(20 - len(fmtd))

        return fmtd

    #FIXME: Update these to correspond with the new error codes
    def parse_result(self, num_ints, buff):
        parsed = {'ints':[], 'buffer':"", 'error': 'No Error', 'is_error': False}

        status_code = self.status
        complete_status = self.complete_status & 0b01111111

        #Check for a command stream layer error
        if self.complete_status == 254:
            parsed['error'] = self.result
            parsed['status'] = self.complete_status
            parsed['is_error'] = True

            return parsed

        parsed['status'] = complete_status

        #Check if the module was not found
        if self.complete_status == 0xFF:
            parsed['error'] = 'Module at address ' + str(self.addr) + ' not found.'
            parsed['is_error'] = True
            return parsed

        #Check for protocol defined errors
        if not complete_status & (1<<6):
            #This is a protocol defined error since the App Defined bit is not set
            if status_code == 0:
                parsed['error'] = 'Module Busy'
            elif status_code == 1:
                parsed['error'] = 'Checksum Error'
            elif status_code == 2:
                parsed['error'] = 'Unsupported Command'
            else:
                parsed['error'] = 'Unrecognized MIB status code'

            parsed['is_error'] = True
            return parsed

        #Otherwise, parse the results according to the type information given
        size = len(self.result)

        if size < 2*num_ints:
            raise RPCException(300, 'Return value too short to unpack : %s' % self.result)
        elif buff == False and size != 2*num_ints:
            raise RPCException(301, 'Return value does not match return type: %s' % self.result)

        for i in xrange(0, num_ints):
            low = ord(self.result[2*i])
            high = ord(self.result[2*i + 1])
            parsed['ints'].append((high << 8) | low)

        if buff:
            parsed['buffer'] = self.result[2*num_ints:]

        return parsed



# This file is adapted from python code released by WellDone International
# under the terms of the LGPLv3.  WellDone International's contact information is
# info@welldone.org
# http://welldone.org
#
# Modifications to this file from the original created at WellDone International 
# are copyright Arch Systems Inc.

#typeinfo.py
#Basic routines for converting information from string or other binary
#formats to python types and for displaying those types in supported 
#formats
#TODO: 
#- Extend the type system to use a recursive parser to allow complex
#  types to be built from other complex types.

import pkg_resources
from iotile.core.exceptions import *
import types
import os.path
import imp

#Start working on recursive parser
#import pyparsing
#symbolchars = pyparsing.Regex('[_a-zA-Z][_a-zA-Z0-9]*')
#typename = pyparsing.Word(symbolchars)

#simpletype = typename
#complextype = pyparsing.Forward()

#typelist = pyparsing.delimitedList(simpletype | complextype, ',')
#complextype << typename + pyparsing.Literal('(').suppress() + typelist + pyparsing.Literal(')').suppress()

#statement = 

class TypeSystem(object):
    """
    TypeSystem permits the inspection of defined types and supports
    converted string and binary values to and from these types.
    """

    def __init__(self, *args):
        """
        Create a TypeSystem by importing all of the types defined in modules passed
        as arguments to this function.  Each module is imported using 
        """

        self.interactive = False
        self.known_types = {}
        self.type_factories = {}

        for arg in args:
            self.load_type_module(arg)

    def convert_to_type(self, value, type, **kwargs):
        """
        Convert value to type 'type'

        If the conversion routine takes various kwargs to 
        modify the conversion process, **kwargs is passed
        through to the underlying conversion function
        """

        if isinstance(value, bytearray):
            return self.convert_from_binary(value, type, **kwargs)

        typeobj = self.get_type(type)

        conv = typeobj.convert(value, **kwargs)
        return conv

    def convert_from_binary(self, binvalue, type, **kwargs):
        """
        Convert binary data to type 'type'.

        'type' must have a convert_binary function.  If 'type'
        supports size checking, the size function is called to ensure
        that binvalue is the correct size for deserialization  
        """

        size = self.get_type_size(type)
        if size > 0 and len(binvalue) != size:
            raise ArgumentError("Could not convert type from binary since the data was not the correct size", required_size=size, actual_size=len(binvalue), type=type)

        typeobj = self.get_type(type)

        if not hasattr(typeobj, 'convert_binary'):
            raise ArgumentError("Type does not support conversion from binary", type=type)

        return typeobj.convert_binary(binvalue, **kwargs)

    def get_type_size(self, type):
        """
        Get the size of this type for converting a hex string to the
        type. Return 0 if the size is not known.
        """

        typeobj = self.get_type(type)

        if hasattr(typeobj, 'size'):
            return typeobj.size()

        return 0

    def format_value(self, value, type, format=None, **kwargs):
        """
        Convert value to type and format it as a string

        type must be a known type in the type system and format,
        if given, must specify a valid formatting option for the
        specified type.
        """

        typed_val = self.convert_to_type(value, type, **kwargs)
        typeobj = self.get_type(type)

        #Allow types to specify default formatting functions as 'default_formatter'
        #otherwise if not format is specified, just convert the value to a string
        if format is None:
            if hasattr(typeobj, 'default_formatter'):
                format_func = getattr(typeobj, 'default_formatter')
                return format_func(typed_val, **kwargs)

            return str(typed_val)

        formatter = "format_%s" % str(format)
        if not hasattr(typeobj, formatter):
            raise ArgumentError("Unknown format for type", type=type, format=format, formatter_function=formatter)

        format_func = getattr(typeobj, formatter)
        return format_func(typed_val, **kwargs)

    def _validate_type(self, typeobj):
        """
        Validate that all required type methods are implemented.

        At minimum a type must have:
        - a convert() or convert_binary() function
        - a default_formatter() function

        Raises an ArgumentError if the type is not valid
        """

        if not (hasattr(typeobj, "convert") or hasattr(typeobj, "convert_binary")):
            raise ArgumentError("type is invalid, does not have convert or convert_binary function", type=typeobj, methods=dir(typeobj))

        if not hasattr(typeobj, "default_formatter"):
            raise ArgumentError("type is invalid, does not have default_formatter function", type=typeobj, methods=dir(typeobj))

    def is_known_type(self, type):
        """
        Check if type is known to the type system.

        Returns boolean indicating if type is known.
        """

        if not isinstance(type, basestring):
            raise ArgumentError("type must be a string naming a known type", type=type)

        if type in self.known_types:
            return True

        return False

    def split_type(self, typename):
        """
        Given a potentially complex type, split it into its base type and specializers
        """

        name = self._canonicalize_type(typename)
        if '(' not in name:
            return name, False, []

        base,sub = name.split('(')
        if len(sub) == 0 or sub[-1] != ')':
            raise ArgumentError("syntax error in complex type, no matching ) found", passed_type=typename, basetype=base, subtype_string=sub)
        
        sub = sub[:-1]

        subs = sub.split(',')
        return base, True, subs

    def instantiate_type(self, typename, base, subtypes):
        """
        Instantiate a complex type
        """

        if base not in self.type_factories:
            raise ArgumentError("unknown complex base type specified", passed_type=typename, base_type=base)

        BaseType = self.type_factories[base]

        #Make sure all of the subtypes are valid
        for s in subtypes:
            try:
                self.get_type(s)
            except IOTileException as e:
                raise ArgumentError("could not instantiate subtype for complex type", passed_type=typename, sub_type=s, error=e)

        typeobj = BaseType.Build(*subtypes, type_system=self)
        self.inject_type(typename, typeobj)

    def _canonicalize_type(self, typename):
        return typename.replace(' ', '')

    def get_type(self, typename):
        """
        Return the type object corresponding to a type name.
        """

        typename = self._canonicalize_type(typename)

        type, is_complex, subtypes = self.split_type(typename)
        if not self.is_known_type(typename):
            if is_complex:
                self.instantiate_type(typename, type, subtypes)
            else:
                raise ArgumentError("get_type called on unknown type", type=typename)

        return self.known_types[typename]

    def is_known_format(self, type, format):
        """
        Check if format is known for given type.

        Returns boolean indicating if format is valid for the specified type.
        """

        typeobj = self.get_type(type)

        formatter = "format_%s" % str(format)
        if not hasattr(typeobj, formatter):
            return False

        return True

    def _is_factory(self, typeobj):
        """
        Determine if typeobj is a factory for producing complex types
        """

        if hasattr(typeobj, 'Build'):
            return True

        return False

    def format_return_value(self, function, value):
        """
        Format the return value of a function based on the annotated type information
        """

        return self.format_value(value, function.retval_typename, function.retval_formatter)

    def inject_type(self, name, typeobj):
        """
        Given a module-like object that defines a type, add it to our type system so that
        it can be used with the iotile tool and with other annotated API functions.
        """

        name = self._canonicalize_type(name)
        base,is_complex,subs = self.split_type(name)

        if self.is_known_type(name):
            raise ArgumentError("attempting to inject a type that is already defined", type=name)

        if (not is_complex) and self._is_factory(typeobj):
            if name in self.type_factories:
                raise ArgumentError("attempted to inject a complex type factory that is already defined", type=name)
            self.type_factories[name] = typeobj
        else:
            self._validate_type(typeobj)
            self.known_types[name] = typeobj

        if not hasattr(typeobj, "default_formatter"):
            raise ArgumentError("type is invalid, does not have default_formatter function", type=typeobj, methods=dir(typeobj))

    def load_type_module(self, module, verbose=False):
        """
        Given a module that contains a list of some types find all symbols in the module that 
        do not start with _ and attempt to import them as types.
        """

        for name in filter(lambda x: not x.startswith('_'), dir(module)):
            typeobj = getattr(module, name)

            try:
                self.inject_type(name, typeobj)
            except ArgumentError:
                pass

    def load_external_types(self, path, verbose=False):
        """
        Given a path to a python package or module, load that module, search for all defined variables
        inside of it that do not start with _ or __ and inject them into the type system.  If any of the
        types cannot be injected, silently ignore them unless verbose is True.  If path points to a module
        it should not contain the trailing .py since this is added automatically by the python import system
        """

        d,p = os.path.split(path)

        try:
            fileobj,pathname,description = imp.find_module(p, [d])
            mod = imp.load_module(p, fileobj, pathname, description)
        except ImportError as e:
            raise ArgumentError("could not import module in order to load external types", module_path=path, parent_directory=d, module_name=p, error=str(e))

        self.load_type_module(mod, verbose)

        #TODO add checking for types that could not be injected and report them

    def load_all_components(self):
        """
        Allow all registered iotile components that have associated type libraries to 
        add themselves to the global type system.
        """

        # Find all of the registered IOTile components and see if we need to add any type libraries for them
        from iotile.core.dev.registry import ComponentRegistry

        reg = ComponentRegistry()
        modules = reg.list_components()

        typelibs = reduce(lambda x,y: x+y, [reg.find_component(x).type_packages() for x in modules], [])
        for lib in typelibs:
            if lib.endswith('.py'):
                lib = lib[:-2]

            self.load_external_types(lib)

        #Also search through install distributions for type libraries
        for entry in pkg_resources.iter_entry_points('iotile.type_package'):
            mod = entry.load()
            self.load_type_module(mod) 


#In order to support function annotations that must be resolved to types when modules
#are imported, create a default TypeSystem object that is used globally to store type
#information

type_system = TypeSystem(types)

#Allow iotile plugins to register themselves and include their types
type_system.load_all_components()

def iprint(stringable):
    """
    A simple function to only print text if in an interactive session.
    """

    if type_system.interactive:
        print str(stringable)



# This file is copyright Arch Systems, Inc.  
# Except as otherwise provided in the relevant LICENSE file, all rights are reserved.

import itertools
from collections import namedtuple
import json
import os.path
from iotile.core.exceptions import *
from semver import SemanticVersion, SemanticVersionRange

ReleaseStep = namedtuple('ReleaseStep', ['provider', 'args'])


class IOTile(object):
    """
    IOTile

    A python representation of an IOTile module allowing you to inspect the products
    that its build produces and include it as a dependency in another build process.
    """

    def __init__(self, folder):
        self.folder = folder
        self.filter_prods = False

        self._load_settings()

    def _load_settings(self):
        modfile = os.path.join(self.folder, 'module_settings.json')

        try:
            with open(modfile, "r") as f:
                settings = json.load(f)
        except IOError:
            raise EnvironmentError("Could not load module_settings.json file, make sure this directory is an IOTile component", path=self.folder)

        if 'module_name' in settings:
            modname = settings['module_name']
        if 'modules' not in settings or len(settings['modules']) == 0:
            raise DataError("No modules defined in module_settings.json file")
        elif len(settings['modules']) > 1:
            raise DataError("Mulitple modules defined in module_settings.json file", modules=settings['modules'].keys())
        else:
            #TODO: Remove this other option once all tiles have been converted to have their own name listed out
            modname = settings['modules'].keys()[0]
        
        if modname not in settings['modules']:
            raise DataError("Module name does not correspond with an entry in the modules directory", name=modname, modules=settings['modules'].keys())
        
        modsettings = settings['modules'][modname]
        architectures = {}
        if 'architectures' in settings:
            architectures = settings['architectures']

        self.settings = modsettings

        #Name is converted to all lowercase to canonicalize it
        prepend = ''
        if 'domain' in modsettings:
            prepend = modsettings['domain'].lower() + '/'

        key = prepend + modname.lower()

        #Copy over some key properties that we want easy access to
        self.name = key
        self.unique_id = key.replace('/', '_')
        self.short_name = modname

        self.full_name = "Undefined"
        if "full_name" in self.settings:
            self.full_name = self.settings['full_name']

        #FIXME: make sure this is a list
        self.authors = []
        if "authors" in self.settings:
            self.authors = self.settings['authors']

        self.version = "0.0.0"
        if "version" in self.settings:
            self.version = self.settings['version']

        self.parsed_version = SemanticVersion.FromString(self.version)

        #Load all of the build products that can be created by this IOTile
        self.products = modsettings.get('products', {})

        #Load in the release information telling us how to release this component
        release_steps = modsettings.get('release_steps', [])
        self.release_steps = []
        self.can_release = False

        for step in release_steps:
            if 'provider' not in step:
                raise DataError("Invalid release step that did not have a provider key", step=step)

            parsed_step = ReleaseStep(provider=step['provider'], args=step.get('args', {}))
            self.release_steps.append(parsed_step)

        if len(self.release_steps) > 0:
            self.can_release = True

        self.dependency_versions = {}

        #If this is a release IOTile component, check for release information
        if 'release' in settings and settings['release'] is True:
            self.release = True

            if 'release_date' not in settings:
                raise DataError("Release mode IOTile component did not include a release date")

            import dateutil.parser
            self.release_date = dateutil.parser.parse(settings['release_date'])
            self.output_folder = self.folder

            if 'dependency_versions' in settings:
                self.dependency_versions = {x: SemanticVersion.FromString(y) for x,y in settings['dependency_versions'].iteritems()}
        else:
            self.release = False
            self.output_folder = os.path.join(self.folder, 'build', 'output')

            #If this tile is a development tile and it has been built at least one, add in a release date
            #from the last time it was built
            if os.path.exists(os.path.join(self.output_folder, 'module_settings.json')):
                release_settings = os.path.join(self.output_folder, 'module_settings.json')

                with open(release_settings, 'rb') as f:
                    release_dict = json.load(f)

                import dateutil.parser
                self.release_date = dateutil.parser.parse(release_dict['release_date'])
            else:
                self.release_date = None

        #Find all of the things that this module could possibly depend on
        #Dependencies include those defined in the module itself as well as
        #those defined in architectures that are present in the module_settings.json
        #file.
        self.dependencies = []

        archs_with_deps = [y['depends'].iteritems() for x,y in architectures.iteritems() if 'depends' in y]
        if 'depends' in self.settings:
            archs_with_deps.append(self.settings['depends'].iteritems())

        #Also search through overlays to architectures that are defined in this module_settings.json file
        #and see if those overlays contain dependencies.
        if 'overlays' in self.settings:
            for overlay_arch in self.settings['overlays'].itervalues():
                if 'depends' in overlay_arch:
                    archs_with_deps.append(overlay_arch['depends'].iteritems())

        found_deps = set()
        for dep, _ in itertools.chain(*archs_with_deps):
            name, _, version = dep.partition(',')
            unique_id = name.lower().replace('/', '_')

            version = version.strip()
            if version == '':
                version = "*"

            ver_range = SemanticVersionRange.FromString(version)

            depdict = {
                'name': name,
                'unique_id': unique_id,
                'required_version': ver_range,
                'required_version_string': version
            }

            if name not in found_deps:
                self.dependencies.append(depdict)

            found_deps.add(name)

        #Setup our support package information
        self.support_distribution = "iotile_support_{0}_{1}".format(self.short_name, self.parsed_version.major)
        self.support_wheel = "{0}-{1}-py2-none-any.whl".format(self.support_distribution, self.version)
        self.has_wheel = False

        if len(self.proxy_modules()) > 0 or len(self.proxy_plugins()) > 0 or len(self.type_packages()) > 0:
            self.has_wheel = True

    def include_directories(self):
        """
        Return a list of all include directories that this IOTile could provide other tiles
        """

        #Only return include directories if we're returning everything or we were asked for it 
        if self.filter_prods and 'include_directories' not in self.desired_prods:
            return []

        if 'include_directories' in self.products:
            if self.release:
                joined_dirs = [os.path.join(self.output_folder, 'include', *x) for x in self.products['include_directories']]
            else:
                joined_dirs = [os.path.join(self.output_folder, *x) for x in self.products['include_directories']]
            return joined_dirs

        return []

    def libraries(self):
        """
        Return a list of all libraries produced by this IOTile that could be provided to other tiles
        """

        libs = [x[0] for x in self.products.iteritems() if x[1] == 'library']

        if self.filter_prods:
            libs = [x for x in libs if x in self.desired_prods]

        badlibs = filter(lambda x: not x.startswith('lib'), libs)
        if len(badlibs) > 0:
            raise DataError("A library product was listed in a module's products without the name starting with lib", bad_libraries=badlibs)

        #Remove the prepended lib from each library name
        return [x[3:] for x in libs]

    def type_packages(self):
        """
        Return a list of the python type packages that are provided by this tile
        """

        libs = [x[0] for x in self.products.iteritems() if x[1] == 'type_package']

        if self.filter_prods:
            libs = [x for x in libs if x in self.desired_prods]

        libs = [os.path.join(self.folder, x) for x in libs]

        return libs

    def linker_scripts(self):
        """
        Return a list of the linker scripts that are provided by this tile
        """

        ldscripts = [x[0] for x in self.products.iteritems() if x[1] == 'linker_script']

        if self.filter_prods:
            ldscripts = [x for x in ldscripts if x in self.desired_prods]

        # Now append the whole path so that the above comparison works based on the name of the product only
        ldscripts = [os.path.join(self.output_folder, 'linker', x) for x in ldscripts]
        return ldscripts

    def proxy_modules(self):
        """
        Return a list of the python proxy modules that are provided by this tile
        """

        libs = [x[0] for x in self.products.iteritems() if x[1] == 'proxy_module']

        if self.filter_prods:
            libs = [x for x in libs if x in self.desired_prods]

        libs = [os.path.join(self.folder, x) for x in libs]
        return libs

    def proxy_plugins(self):
        """
        Return a list of the python proxy plugins that are provided by this tile
        """

        libs = [x[0] for x in self.products.iteritems() if x[1] == 'proxy_plugin']

        if self.filter_prods:
            libs = [x for x in libs if x in self.desired_prods]

        libs = [os.path.join(self.folder, x) for x in libs]
        return libs

    def firmware_images(self):
        """
        Return a list of the python proxy plugins that are provided by this tile
        """

        libs = [x[0] for x in self.products.iteritems() if x[1] == 'firmware_image']

        if self.filter_prods:
            libs = [x for x in libs if x in self.desired_prods]

        libs = [os.path.join(self.output_folder, x) for x in libs]
        return libs

    def tilebus_definitions(self):
        """
        Return a list of all tilebus definitions that this IOTile could provide other tiles
        """

        #Only return include directories if we're returning everything or we were asked for it 
        if self.filter_prods and 'tilebus_definitions' not in self.desired_prods:
            return []

        if 'tilebus_definitions' in self.products:
            if self.release:
                #For released tiles, all of the tilebus definitions are copies to the same directory
                joined_dirs = [os.path.join(self.output_folder, 'tilebus', os.path.basename(os.path.join(*x))) for x in self.products['tilebus_definitions']]
            else:
                joined_dirs = [os.path.join(self.folder, 'firmware', 'src', *x) for x in self.products['tilebus_definitions']]
            return joined_dirs

        return []

    def library_directories(self):
        libs = self.libraries()

        if len(libs) > 0:
            return [os.path.join(self.output_folder)]

        return []

    def filter_products(self, desired_prods):
        """
        When asked for a product that this iotile produces, filter only those on this list
        """

        self.filter_prods = True
        self.desired_prods = set(desired_prods)

    def path(self):
        return self.folder



# This file is adapted from python code released by WellDone International
# under the terms of the LGPLv3.  WellDone International's contact information is
# info@welldone.org
# http://welldone.org
#
# Modifications to this file from the original created at WellDone International 
# are copyright Arch Systems Inc.

#map.py
#a complex type wrapping a python dictionary

class map(object):
	def __init__(self, keytype, valuetype, **kwargs):
		
		self.keytype = keytype
		self.valuetype = valuetype
		self.type_system = kwargs['type_system']

	@staticmethod
	def Build(*types, **kwargs):
		if len(types) != 2:
			raise ValueError("map must be created with 2 arguments, a keytype and a valuetype")
		
		return map(types[0], types[1], **kwargs)

	def convert(self, value, **kwargs):
		if isinstance(value, dict):
			return value

		raise ValueError("Converting to map from string not yet supported")

	def default_formatter(self, value, **kwargs):
		forms = []
		for key,val in value.iteritems():
			keyform = self.type_system.format_value(key, self.keytype)
			valform = self.type_system.format_value(val, self.valuetype)
			forms.append("%s: %s" % (keyform, valform))

		return "\n".join(forms)


# This file is adapted from python code released by WellDone International
# under the terms of the LGPLv3.  WellDone International's contact information is
# info@welldone.org
# http://welldone.org
#
# Modifications to this file from the original created at WellDone International 
# are copyright Arch Systems Inc.

#config.py

import os
import json

class ConfigFile:
	"""
	A simple wrapper around a json config file with a local and global version
	the local version always overrides the global version.
	"""

	def __init__(self, base_dir, name, require_local=False):
		localpath = os.path.join(base_dir, name +'.local.json')
		globalpath = os.path.join(base_dir, name + '.global.json')

		conf = {}
		localconf = None
		try:
			with open(globalpath, 'r') as f:
				conf = json.load(f)
		except IOError:
			pass #global file can not exist

		try:
			with open(localpath, 'r') as f:
				localconf = json.load(f)
		except IOError:
			if require_local:
				raise ValueError('Local settings file does not exist and is required: %s' % name +'.local.json')

		if localconf is not None:
			merge_dicts(conf, localconf)

		self.conf = conf

	def __getitem__(self, path):
		"""
		Given a path specifying subdictionaries in this config dictionary,
		with each key separated by a /, return the value indicated.  It must
		exist.
		"""

		keys = path.split('/')

		val = self.conf[keys[0]]

		for i in xrange(1, len(keys)):
			val = val[keys[i]]

		return val


#from http://stackoverflow.com/questions/7204805/python-dictionaries-of-dictionaries-merge
def merge_dicts(a, b):
    "merges b into a"

    for key in b:
        if key in a:
            if isinstance(a[key], dict) and isinstance(b[key], dict):
                merge_dicts(a[key], b[key])
            else:
            	a[key] = b[key]
        else:
            a[key] = b[key]
    
    return a


# This file is copyright Arch Systems, Inc.  
# Except as otherwise provided in the relevant LICENSE file, all rights are reserved.

from iotile.core.utilities.kvstore_sqlite import SQLiteKVStore
from iotile.core.utilities.kvstore_json import JSONKVStore
from iotile.core.exceptions import *
import json
import os.path
from iotileobj import IOTile
import pkg_resources
import imp

MISSING = object()


class ComponentRegistry(object):
    """
    ComponentRegistry

    A mapping of all of the installed components on this system that can
    be used as build dependencies and where they are located.  Also used
    to manage iotile plugins.
    """

    BackingType = SQLiteKVStore
    BackingFileName = 'component_registry.db'

    def __init__(self):
        self.kvstore = self.BackingType(self.BackingFileName, respect_venv=True)
        self.plugins = {}

        for entry in pkg_resources.iter_entry_points('iotile.plugin'):
            plugin = entry.load()
            links = plugin()
            for name, value in links:
                self.plugins[name] = value

    @classmethod
    def SetBackingStore(cls, backing):
        """Set the global backing type used by the ComponentRegistry from this point forward

        This function must be called before any operations that use the registry are initiated
        otherwise they will work from different registries that will likely contain different data
        """

        if backing not in ['json', 'sqlite']:
            raise ArgumentError("Unknown backing store type that is not json or sqlite", backing=backing)

        if backing == 'json':
            cls.BackingType = JSONKVStore
            cls.BackingFileName = 'component_registry.json'
        else:
            cls.BackingType = SQLiteKVStore
            cls.BackingFileName = 'component_registry.db'

    def add_component(self, component):
        """
        Register a component with ComponentRegistry. 

        Component must be a buildable object with a module_settings.json file that
        describes its name and the domain that it is part of.
        """

        tile = IOTile(component)
        value = os.path.normpath(os.path.abspath(component))

        self.kvstore.set(tile.name, value)

    def get_component(self, component):
        try:
            comp_path = self.kvstore.get(component)
        except KeyError:
            raise ArgumentError("Could not find component by name", component=component)

        return IOTile(comp_path)

    def list_plugins(self):
        """
        List all of the plugins that have been registerd for the iotile program on this computer
        """

        vals = self.plugins.items()

        return {x: y for x, y in vals}

    def find_component(self, key, domain=""):
        try:
            if domain is not "":
                key = domain.lower() + '/' + key.lower()

            return IOTile(self.kvstore.get(key))
        except KeyError:
            raise ArgumentError("Unknown component name", name=key)

    def remove_component(self, key):
        """Remove component from registry
        """

        return self.kvstore.remove(key)

    def clear_components(self):
        """Clear all of the registered components
        """

        for key in self.list_components():
            self.remove_component(key)

    def clear(self):
        """Clear all data from the registry
        """

        self.kvstore.clear()

    def list_components(self):
        """List all of the registered component names
        """

        items = self.kvstore.get_all()

        return [x[0] for x in items if not x[0].startswith('config:')]

    def set_config(self, key, value):
        """Set a persistent config key to a value, stored in the registry

        Args:
            key (string): The key name
            value (string): The key value
        """

        keyname = "config:" + key

        self.kvstore.set(keyname, value)

    def get_config(self, key, default=MISSING):
        """Get the value of a persistent config key from the registry

        If no default is specified and the key is not found ArgumentError is raised.

        Args:
            key (string): The key name to fetch
            default (string): an optional value to be returned if key cannot be found

        Returns:
            string: the key's value
        """

        keyname = "config:" + key

        try:
            return self.kvstore.get(keyname)
        except KeyError:
            if default is MISSING:
                raise ArgumentError("No config value found for key", key=key)

            return default

    def clear_config(self, key):
        """Remvoe a persistent config key from the registry

        Args:
            key (string): The key name
        """

        keyname = "config:" + key
        self.kvstore.remove(keyname)



import pkg_resources
import json
import traceback
from adapter import DeviceAdapter
from iotile.core.exceptions import ArgumentError
from iotile.core.hw.virtual.virtualdevice import RPCInvalidIDError, TileNotFoundError, RPCNotFoundError

class VirtualDeviceAdapter(DeviceAdapter):
    """Callback based adapter that gives access to one or more virtual devices
    
    The adapter is created and serves access to the virtual_devices that are
    found by name in the entry_point group iotile.virtual_device.

    Args:
        port (string): A port description that should be in the form of
            device_name1@<optional_config_json1;device_name2@optional_config_json2
    """

    def __init__(self, port):
        super(VirtualDeviceAdapter, self).__init__()

        devs = port.split(';')
        loaded_devs = {}

        for dev in devs:
            name, sep, config = dev.partition('@')

            if len(config) == 0:
                config = None

            loaded_dev = self._load_device(name, config)
            loaded_devs[loaded_dev.iotile_id] = loaded_dev

        self.devices = loaded_devs
        self.connections = {}

    def _load_device(self, name, config):
        if config is None:
            config_dict = {}
        else:
            with open(config, "rb") as conf:
                data = json.load(conf)

            if 'device' not in data:
                raise ArgumentError("Invalid configuration file passed to VirtualDeviceAdapter", device_name=name, config_path=config, missing_key='device')

            config_dict = data['device']

        seen_names = []
        for entry in pkg_resources.iter_entry_points('iotile.virtual_device'):
            if entry.name == name:
                device_factory = entry.load()
                return device_factory(config_dict)

            seen_names.append(entry.name)

        raise ArgumentError("Could not find virtual_device by name", name=name, known_names=seen_names)

    def connect_async(self, connection_id, connection_string, callback):
        """Asynchronously connect to a device

        Args:
            connection_id (int): A unique identifer that will refer to this connection
            connection_string (string): A DeviceAdapter specific string that can be used to connect to
                a device using this DeviceAdapter.
            callback (callable): A function that will be called when the connection attempt finishes as
                callback(conection_id, adapter_id, success: bool, failure_reason: string or None)
        """

        id_number = int(connection_string)
        if id_number not in self.devices:
            if callback is not None:
                callback(connection_id, self.id, False, "could not find device to connect to")
            return
        
        if id_number in [x.iotile_id for x in self.connections.itervalues()]:
            if callback is not None:
                callback(connection_id, self.id, False, "device was already connected to")
            return

        dev = self.devices[id_number]

        self.connections[connection_id] = dev

        if callback is not None:
            callback(connection_id, self.id, True, "")

    def disconnect_async(self, conn_id, callback):
        """Asynchronously disconnect from a connected device

        Args:
            conn_id (int): A unique identifer that will refer to this connection
            callback (callback): A callback that will be called as
                callback(conn_id, adapter_id, success, failure_reason)
        """

        if conn_id not in self.connections:
            if callback is not None:
                callback(conn_id, self.id, False, "device had no active connection")
            return

        dev = self.connections[conn_id]
        del self.connections[conn_id]

        if callback is not None:
            callback(conn_id, self.id, True, "")

    def _open_rpc_interface(self, conn_id, callback):
        """Open the RPC interface on a device

        Args:
            conn_id (int): A unique identifer that will refer to this connection
            callback (callback): A callback that will be called as
                callback(conn_id, adapter_id, success, failure_reason)
        """

        if conn_id not in self.connections:
            if callback is not None:
                callback(conn_id, self.id, False, "device had no active connection")
            return

        dev = self.connections[conn_id]
        dev.open_rpc_interface()

        if callback is not None:
            callback(conn_id, self.id, True, "")

    def _open_streaming_interface(self, conn_id, callback):
        """Open the streaming interface on a device

        Args:
            conn_id (int): A unique identifer that will refer to this connection
            callback (callback): A callback that will be called as
                callback(conn_id, adapter_id, success, failure_reason)
        """

        if conn_id not in self.connections:
            if callback is not None:
                callback(conn_id, self.id, False, "device had no active connection")
            return

        dev = self.connections[conn_id]
        reports = dev.open_streaming_interface()

        if callback is not None:
            callback(conn_id, self.id, True, "")

        for report in reports:
            self._trigger_callback('on_report', conn_id, report)

    def _open_tracing_interface(self, conn_id, callback):
        """Open the tracing interface on a device

        Args:
            conn_id (int): A unique identifer that will refer to this connection
            callback (callback): A callback that will be called as
                callback(conn_id, adapter_id, success, failure_reason)
        """

        if conn_id not in self.connections:
            if callback is not None:
                callback(conn_id, self.id, False, "device had no active connection")
            return

        dev = self.connections[conn_id]
        traces = dev.open_tracing_interface()

        if callback is not None:
            callback(conn_id, self.id, True, "")

        for trace in traces:
            self._trigger_callback('on_trace', conn_id, trace)

    def _open_script_interface(self, conn_id, callback):
        """Open the script interface on a device

        Args:
            conn_id (int): A unique identifer that will refer to this connection
            callback (callback): A callback that will be called as
                callback(conn_id, adapter_id, success, failure_reason)
        """

        if conn_id not in self.connections:
            if callback is not None:
                callback(conn_id, self.id, False, "device had no active connection")
            return

        dev = self.connections[conn_id]
        dev.open_script_interface()

        if callback is not None:
            callback(conn_id, self.id, True, "")

    def send_rpc_async(self, conn_id, address, rpc_id, payload, timeout, callback):
        """Asynchronously send an RPC to this IOTile device

        Args:
            conn_id (int): A unique identifer that will refer to this connection
            address (int): the addres of the tile that we wish to send the RPC to
            rpc_id (int): the 16-bit id of the RPC we want to call
            payload (bytearray): the payload of the command
            timeout (float): the number of seconds to wait for the RPC to execute
            callback (callable): A callback for when we have finished the RPC.  The callback will be called as" 
                callback(connection_id, adapter_id, success, failure_reason, status, payload)
                'connection_id': the connection id
                'adapter_id': this adapter's id
                'success': a bool indicating whether we received a response to our attempted RPC
                'failure_reason': a string with the reason for the failure if success == False
                'status': the one byte status code returned for the RPC if success == True else None
                'payload': a bytearray with the payload returned by RPC if success == True else None
        """

        if conn_id not in self.connections:
            if callback is not None:
                callback(conn_id, self.id, False, 'Device is not in connected state', None, None)
            return

        dev = self.connections[conn_id]

        status = (1 << 6)
        try:
            response = dev.call_rpc(address, rpc_id, str(payload))
            if len(response) > 0:
                status |= (1 << 7)
        except (RPCInvalidIDError, RPCNotFoundError):
            status = 2
            response = ""
        except TileNotFoundError:
            status = 0xFF
            response = ""
        except Exception:
            #Don't allow exceptions or we will deadlock
            status = 3
            response = ""

            print("*** EXCEPTION OCCURRED IN RPC ***")
            traceback.print_exc()
            print("*** END EXCEPTION ***")

        response = bytearray(response)
        callback(conn_id, self.id, True, "", status, response)

    def periodic_callback(self):
        pass

    def stop_sync(self):
        pass



# This file is adapted from python code released by WellDone International
# under the terms of the LGPLv3.  WellDone International's contact information is
# info@welldone.org
# http://welldone.org
#
# Modifications to this file from the original created at WellDone International 
# are copyright Arch Systems Inc.

#MIB Proxy Objects

from iotile.core.hw.commands import RPCCommand
from iotile.core.hw.exceptions import *
from iotile.core.utilities.typedargs import return_type, annotated, param, context
from time import sleep
from iotile.core.utilities.packed import unpack
import struct
from iotile.core.exceptions import *

class TileBusProxyObject (object):
	def __init__(self, stream, address):
		self.stream = stream
		self.addr = address
		self._config_manager = ConfigManager(self)

	@classmethod
	def ModuleName(cls):
		return 'NO APP'

	@annotated
	def config_manager(self):
		return self._config_manager

	def rpc(self, feature, cmd, *args, **kw):
		"""
		Send an RPC call to this module, interpret the return value
		according to the result_type kw argument.  Unless raise keyword
		is passed with value False, raise an RPCException if the command
		is not successful.
		"""

		status, payload = self.stream.send_rpc(self.addr, feature, cmd, *args, **kw)

		unpack_flag = False
		if "result_type" in kw:
			res_type = kw['result_type']
		elif "result_format" in kw:
			unpack_flag = True
			res_type = (0, True)
		else:
			res_type = (0, False)

		try:
			res = self._parse_rpc_result(status, payload, *res_type)
			if unpack_flag:
				return unpack("<%s" % kw["result_format"], res['buffer'])

			return res
		except ModuleBusyError:
			pass

		if "retries" not in kw:
			kw['retries'] = 10

		#Sleep 100 ms and try again unless we've exhausted our retry attempts
		if kw["retries"] > 0:
			kw['retries'] -= 1

			sleep(0.1)
			return self.rpc(feature, cmd, *args, **kw)

	@return_type("basic_dict")
	def status(self):
		"""
		Query the status of an IOTile including its name and version
		"""

		hw_type, name, major, minor, patch, status = self.rpc(0x00, 0x04, result_format="H6sBBBB")

		status = {
			'hw_type': hw_type,
			'name': name,
			'version': (major, minor, patch),
			'status': status
		}

		return status

	@param("wait", "float", desc="Time to wait after reset for tile to boot up to a usable state")
	def reset(self, wait=1.0):
		"""
		Immediately reset this tile.
		"""
		try:
			self.rpc(0x00, 0x01)
		except ModuleNotFoundError:
			pass

		sleep(wait)

	@return_type("string")
	def tile_name(self):
		stat = self.status()

		return stat['name']

	@return_type("list(integer)", formatter="compact")
	def tile_version(self):
		stat = self.status()

		return stat['version']

	@return_type("map(string, bool)")
	def tile_status(self):
		"""	
		Get the current status of this tile

		Returns a 
		"""
		stat = self.status()

		flags = stat['status']

		#FIXME: This needs to stay in sync with lib_common: cdb_status.h
		status = {}
		status['debug_mode'] = bool(flags & (1 << 3))
		status['configured'] = bool(flags & (1 << 1))
		status['app_running'] = bool(flags & (1 << 0))
		status['trapped'] = bool(flags & (1 << 2))

		return status


	def _parse_rpc_result(self, status, payload, num_ints, buff):
		"""
		Parse the response of an RPC call into a dictionary with integer and buffer results
		"""

		parsed = {'ints':[], 'buffer':"", 'error': 'No Error', 'is_error': False}
		parsed['status'] = status
		parsed['return_value'] = status & 0b00111111

		#Check for protocol defined errors
		if not status & (1<<6):
			if status == 2:
				raise UnsupportedCommandError(address=self.addr)
			
			raise RPCError("Unknown status code received from RPC call", address=self.addr, status_code=status)


		#Otherwise, parse the results according to the type information given
		size = len(payload)

		if size < 2*num_ints:
			raise RPCError('Return value too short to unpack', expected_minimum_size=2*num_ints, actual_size=size, status_code=status, payload=payload)
		elif buff == False and size != 2*num_ints:
			raise RPCError('Return value was not the correct size', expected_size=2*num_ints, actual_size=size, status_code=status, payload=payload)

		for i in xrange(0, num_ints):
			low = (payload[2*i])
			high = (payload[2*i + 1])
			parsed['ints'].append((high << 8) | low)

		if buff:
			parsed['buffer'] = payload[2*num_ints:]

		return parsed


@context("ConfigManager")
class ConfigManager(object):
	"""
	Manager Proxy for configuration variables on IOTiles

	Handles querying what config variables are defined, setting and getting
	their values.  Note that config variables should not change when application
	code is running so set_config_variables should not be called once an application
	is launched.
	"""

	def __init__(self, parent):
		self._proxy = parent

	@return_type('list(integer)')
	def list_variables(self):
		"""
		List all of the configuration variables defined by this tile
		"""

		offset = 0
		ids = []

		while True:
			resp = self._proxy.rpc(0, 10, offset, result_type=(1, True))
			count = resp['ints'][0]

			if count == 0:
				break

			fmt = "<%dH" % count
			id_chuck = struct.unpack(fmt, resp['buffer'][:2*count])

			ids += id_chuck

			if count != 9:
				break

		return ids

	@return_type("basic_dict")
	@param("id", "integer", desc="Variable ID to describe")
	def describe_variable(self, id):
		"""
		Describe a configuration variable 
		"""

		resp = self._proxy.rpc(0, 11, id, result_type =(2, True))

		err = resp['ints'][0]
		if err != 0:
			raise HardwareError("Error finding config variable by id", id=id, error_code=err)

		addr, id, flags = struct.unpack("<LHH", resp['buffer'])

		maxsize = (flags & ~(1 << 15)) & 0xFFFF
		variable_size = bool(flags >> 15)
		
		info = {
			'address': addr,
			'id': id,
			'max_size': maxsize,
			'variable_size': variable_size
		}

		return info

	@return_type("bytes", "repr")
	@param("id", "integer", desc="Variable ID to fetch")
	def get_variable(self, id):
		"""
		Get value stored in a config variable
		"""

		offset = 0
		resp = self._proxy.rpc(0, 13, id, offset, result_type=(0, True))
		if len(resp['buffer']) == 0:
			return bytearray()

		retval = resp['buffer']

		while len(resp['buffer']) > 0:
			offset = len(retval)
			resp = self._proxy.rpc(0, 13, id, offset, result_type=(0, True))

			retval += resp['buffer']

		return retval

	@param("id", "integer", desc="Variable ID to set")
	@param("value", "bytes", desc="hexadecimal byte value to set")
	def set_variable(self, id, value):
		"""
		Set the value stored in a config variable
		"""

		for offset in xrange(0, len(value), 16):
			remaining = len(value) - offset
			if remaining > 16:
				remaining = 16
			
			resp = self._proxy.rpc(0, 12, id, offset, value[offset:offset+remaining], result_type=(1, False))
			if resp['ints'][0] != 0:
				raise HardwareError("Error setting config variable", id=id, error_code=resp['ints'][0])



import threading
from iotile.core.exceptions import ArgumentError

MISSING = object()

class DeviceAdapter(object):
    """Classes that encapsulate access to IOTile devices over a particular communication channel

    Subclasses of DeviceAdapter implement concrete access channels over which IOTile devices
    can be controlled and send data. Examples area a bluetooth low energy communication channel
    or USB.  In order to fit into the rest of the IOTile tooling systems, only a few functions
    need to be implemented in a DeviceAdapter as shown below.  At its core, DeviceAdapters need
    to be able to connect/discconect from a device, open/close an interface on the device, send RPCs
    and send scripts.

    The interface to a DeviceAdapter is:
    connect_async
    connect_sync

    disconnect_async
    disconnect_sync

    open_interface_async
    open_interface_sync

    close_interface_async
    close_interface_sync

    send_rpc_async
    send_rpc_sync

    send_script_async
    send_script_sync

    periodic_callback
    can_connect

    stop_sync

    Subclasses only need to override the '_async' versions of each call.  The synchronous versions will
    be automatically functional using the '_async' versions provided that the '_async' version not use
    multiprocessing to invoke its callback, i.e. it should use multithreading since the default synchronous
    adapter function needs a shared memory lock. 

    periodic_callback should be a non-blocking callback that is invoked periodically to allow
    the DeviceAdapter to maintain its internal state.

    Additionally you can register callbacks that will be called in the following circumstances:

    on_disconnect: called when the device disconnects unexpectedly
        Signature of on_disconnect should be on_disconnect(adapter_id, connection_id)
    on_report: called when the device has streamed a complete sensor graph report
        Signature of on_report should be on_report(connection_id, report), 
        where report is a subclass of IOTileReport.
    on_scan: called when this device is seen during a scan of this communication channel
        Signature of on_scan should be on_scan(adapter_id, device_info, expiration_interval)
    on_trace: called when the device has received tracing data, which is an unstructured string
        of bytes.  Signature of on_trace should be on_trace(connection_id, trace_data),
        where trace_data is a bytearray.
    """

    def __init__(self):
        self.id = -1

        self.callbacks = {}
        self.callbacks['on_scan'] = set()
        self.callbacks['on_disconnect'] = set()
        self.callbacks['on_report'] = set()
        self.callbacks['on_trace'] = set()
        self.config = {}

    def set_id(self, adapter_id):
        """Set an ID that this adapater uses to identify itself when making callbacks
        """

        self.id = adapter_id

    def set_config(self, name, value):
        """Set a config value for this adapter by name

        Args:
            name (string): The name of the config variable
            value (object): The value of the config variable
        """

        self.config[name] = value

    def get_config(self, name, default=MISSING):
        """Get a config value from this adapter by name

        Args:
            name (string): The name of the config variable
            default (object): The default value to return if config is not found

        Returns:
            object: the value associated with the name

        Raises:
            ArgumentError: if the name is not found and no default is supplied
        """

        res = self.config.get(name, default)
        if res is MISSING:
            raise ArgumentError("Could not find config value by name and no default supplied", name=name)

        return res

    def add_callback(self, name, func):
        """Add a callback when Device events happen

        Args:
            name (str): currently support 'on_scan' and 'on_disconnect'
            func (callable): the function that should be called
        """

        if name not in self.callbacks:
            raise ValueError("Unknown callback name: %s" % name)

        self.callbacks[name].add(func)

    def _trigger_callback(self, name, *args, **kwargs):
        for func in self.callbacks[name]:
            func(*args, **kwargs)

    def connect_async(self, connection_id, connection_string, callback):
        """Asynchronously connect to a device

        Args:
            connection_id (int): A unique identifer that will refer to this connection
            connection_string (string): A DeviceAdapter specific string that can be used to connect to
                a device using this DeviceAdapter.
            callback (callable): A function that will be called when the connection attempt finishes as
                callback(conection_id, adapter_id, success: bool, failure_reason: string or None)
        """

        if callback is not None:
            callback(connection_id, self.id, False, "connect command is not supported in device adapter")

    def connect_sync(self, connection_id, connection_string):
        """Synchronously connect to a device

        Args:
            connection_id (int): A unique identifer that will refer to this connection
            connection_string (string): A DeviceAdapter specific string that can be used to connect to
                a device using this DeviceAdapter.

        Returns:
            dict: A dictionary with two elements
                'success': a bool with the result of the connection attempt
                'failure_reaon': a string with the reason for the failure if we failed
        """

        calldone = threading.Event()
        results = {}

        def connect_done(callback_connid, callback_adapterid, callback_success, failure_reason):
            calldone.set()
            results['success'] = callback_success
            results['failure_reason'] = failure_reason

        self.connect_async(connection_id, connection_string, connect_done)
        calldone.wait()

        return results

    def disconnect_async(self, conn_id, callback):
        """Asynchronously disconnect from a connected device

        Args:
            conn_id (int): A unique identifer that will refer to this connection
            callback (callback): A callback that will be called as
                callback(conn_id, adapter_id, success, failure_reason)
        """

        callback(conn_id, self.id, False, "disconnect command is not supported in device adapter")

    def disconnect_sync(self, conn_id):
        """Synchronously disconnect from a connected device

        Args:
            conn_id (int): A unique identifer that will refer to this connection
        
        Returns:
            dict: A dictionary with two elements
                'success': a bool with the result of the connection attempt
                'failure_reaon': a string with the reason for the failure if we failed
        """

        done = threading.Event()
        result = {}

        def disconnect_done(conn_id, adapter_id, status, reason):
            result['success'] = status
            result['failure_reason'] = reason
            done.set()

        self.disconnect_async(conn_id, disconnect_done)
        done.wait()

        return result

    def open_interface_async(self, conn_id, interface, callback):
        """Asynchronously open an interface to this IOTile device

        interface must be one of (rpc, script, streaming)

        Args:
            interface (string): The interface to open
            conn_id (int): A unique identifer that will refer to this connection 
            callback (callable):  A callback that will be called as
                callback(conn_id, adapter_id, success, failure_reason)
        """

        if interface not in set(["rpc", "script", "streaming", "tracing"]):
            callback(conn_id, self.id, False, "invalid interface name in call to open_interface_async")
            return

        if interface == "rpc":
            self._open_rpc_interface(conn_id, callback)
        elif interface == 'script':
            self._open_script_interface(conn_id, callback)
        elif interface == 'streaming':
            self._open_streaming_interface(conn_id, callback)
        elif interface == 'tracing':
            self._open_tracing_interface(conn_id, callback)
        else:
            callback(conn_id, self.id, False, "interface not supported yet")

    def open_interface_sync(self, conn_id, interface):
        """Asynchronously open an interface to this IOTile device

        interface must be one of (rpc, script, streaming)

        Args:
            interface (string): The interface to open
            conn_id (int): A unique identifer that will refer to this connection

        Returns:
            dict: A dictionary with four elements
                'success': a bool with the result of the connection attempt
                'failure_reaon': a string with the reason for the failure if we failed
        """

        done = threading.Event()
        result = {}

        def open_interface_done(conn_id, adapter_id, status, reason):
            result['success'] = status
            result['failure_reason'] = reason
            done.set()

        self.open_interface_async(conn_id, interface, open_interface_done)
        done.wait()

        return result

    def send_rpc_async(self, conn_id, address, rpc_id, payload, timeout, callback):
        """Asynchronously send an RPC to this IOTile device

        Args:
            conn_id (int): A unique identifer that will refer to this connection
            address (int): the addres of the tile that we wish to send the RPC to
            rpc_id (int): the 16-bit id of the RPC we want to call
            payload (bytearray): the payload of the command
            timeout (float): the number of seconds to wait for the RPC to execute
            callback (callable): A callback for when we have finished the RPC.  The callback will be called as" 
                callback(connection_id, adapter_id, success, failure_reason, status, payload)
                'connection_id': the connection id
                'adapter_id': this adapter's id
                'success': a bool indicating whether we received a response to our attempted RPC
                'failure_reason': a string with the reason for the failure if success == False
                'status': the one byte status code returned for the RPC if success == True else None
                'payload': a bytearray with the payload returned by RPC if success == True else None
        """

        callback(conn_id, self.id, False, 'RPCs are not supported on this adapter', None, None)

    def send_rpc_sync(self, conn_id, address, rpc_id, payload, timeout):
        """Synchronously send an RPC to this IOTile device

        Args:
            conn_id (int): A unique identifer that will refer to this connection
            address (int): the addres of the tile that we wish to send the RPC to
            rpc_id (int): the 16-bit id of the RPC we want to call
            payload (bytearray): the payload of the command
            timeout (float): the number of seconds to wait for the RPC to execute

        Returns:
            dict: A dictionary with four elements
                'success': a bool indicating whether we received a response to our attempted RPC
                'failure_reason': a string with the reason for the failure if success == False
                'status': the one byte status code returned for the RPC if success == True else None
                'payload': a bytearray with the payload returned by RPC if success == True else None
        """

        done = threading.Event()
        result = {}

        def send_rpc_done(conn_id, adapter_id, status, reason, rpc_status, resp_payload):
            result['success'] = status
            result['failure_reason'] = reason
            result['status'] = rpc_status
            result['payload'] = resp_payload

            done.set()

        self.send_rpc_async(conn_id, address, rpc_id, payload, timeout, send_rpc_done)
        done.wait()

        return result

    def send_script_async(self, conn_id, data, progress_callback, callback):
        """Asynchronously send a a script to this IOTile device

        Args:
            conn_id (int): A unique identifer that will refer to this connection
            data (string): the script to send to the device
            progress_callback (callable): A function to be called with status on our progress, called as:
                progress_callback(done_count, total_count)
            callback (callable): A callback for when we have finished sending the script.  The callback will be called as" 
                callback(connection_id, adapter_id, success, failure_reason)
                'connection_id': the connection id
                'adapter_id': this adapter's id
                'success': a bool indicating whether we received a response to our attempted RPC
                'failure_reason': a string with the reason for the failure if success == False
        """

        callback(conn_id, self.id, False, 'Sending scripts is not supported by this device adapter')

    def send_script_sync(self, conn_id, data, progress_callback):
        """Asynchronously send a a script to this IOTile device

        Args:
            conn_id (int): A unique identifer that will refer to this connection
            data (string): the script to send to the device
            progress_callback (callable): A function to be called with status on our progress, called as:
                progress_callback(done_count, total_count)

        Returns:
            dict: a dict with the following two entries set
                'success': a bool indicating whether we received a response to our attempted RPC
                'failure_reason': a string with the reason for the failure if success == False
        """

        done = threading.Event()
        result = {}

        def send_script_done(conn_id, adapter_id, status, reason):
            result['success'] = status
            result['failure_reason'] = reason

            done.set()

        self.send_script_async(conn_id, data, progress_callback, send_script_done)
        done.wait()

        return result


# This file is adapted from python code released by WellDone International
# under the terms of the LGPLv3.  WellDone International's contact information is
# info@welldone.org
# http://welldone.org
#
# Modifications to this file from the original created at WellDone International 
# are copyright Arch Systems Inc.

#shell.py
#Given a command line string, attempt to map it to a function and fill in 
#the parameters based on that function's annotated type information.

from iotile.core.exceptions import *
import annotate
import inspect
import shlex
from typeinfo import type_system
from iotile.core.utilities.rcfile import RCFile
import os.path
import platform
import importlib

posix_lex = platform.system() != 'Windows'

builtin_help = {
	'help': "help [function]: print help information about the current context or a function",
	'back': "back: replace the current context with its parent",
	'quit': "quit: quit the momo shell"
}

def process_kwarg(flag, arg_it):
	flag = flag[2:]
	skip = 0

	#Check if of the form name=value
	if '=' in flag:
		name, value = flag.split('=')
	else:
		name = flag
		value = next(arg_it)
		skip = 1

	return name, value, skip

@annotate.param("package", "path", "exists", desc="Path to the python package containing the types")
@annotate.param("module", "string", desc="The name of the submodule to load from package, if applicable")
def import_types(package, module=None):
	"""
	Add externally defined types from a python package or module

	The MoMo type system is built on the typedargs package, which defines what kinds of types
	can be used for function arguments as well as how those types should be displayed and 
	converted from binary representations or strings.  This function allows you to define external
	types in a separate package and import them into the MoMo type system so that they can be used.
	You might want to do this if you have custom firmware objects that you would like to interact with
	or that are returned in a syslog entry, for example.  

	All objects defined in the global namespace of package (if module is None) or package.module if 
	module is specified that define valid types will be imported and can from this point on be used
	just like any primitive type defined in typedargs itself.  Imported types are indistinguishable 
	from primivtive types like string, integer and path.
	"""

	if module is None:
		path = package
	else:
		path = os.path.join(package, module)

	type_system.load_external_types(path)

def print_dir(context):
	doc = inspect.getdoc(context)
	
	print ""
	print annotate.context_name(context)

	if doc is not None:
		doc = inspect.cleandoc(doc)
		print doc
	
	print "\nDefined Functions:"

	if isinstance(context, dict):
		funs = context.keys()
	else:
		funs = annotate.find_all(context)
	
	for fun in funs:
		fun = find_function(context, fun)
		if isinstance(fun, annotate.BasicContext):
			print " - " + fun._annotated_name
		else:
			print " - " + annotate.get_signature(fun)

		if annotate.short_description(fun) != "":
			print "   " + annotate.short_description(fun) + '\n'
		else:
			print ""

	print "\nBuiltin Functions"
	for bi in builtin_help.values():
		print ' - ' + bi

	print ""

def print_help(context, fname):
	if fname in builtin_help:
		print builtin_help[fname]
		return

	func = find_function(context, fname)
	annotate.print_help(func)

def _do_help(context, line):
	args = line[1:]
	if len(args) == 0:
		print_dir(context)
	elif len(args) == 1:
		print_help(context, args[0])
	else:
		print "Too many arguments:", args
		print "Usage: help [function]"

	return [], True

def deferred_add(add_action):
	"""
	Perform a lazy import of a context so that we don't have a huge initial startup time
	loading all of the modules that someone might want even though they probably only 
	will use a few of them.
	"""

	module, sep, obj = add_action.partition(',')

	mod = importlib.import_module(module)
	if obj == "":
		name, con = annotate.context_from_module(mod)
		return con

	if hasattr(mod, obj):
		return getattr(mod, obj)

	raise ArgumentError("Attempted to import nonexistent object from module", module=module, object=obj)

def find_function(context, funname):
	func = None
	if isinstance(context, dict):
		if funname in context:
			func = context[funname]

			#Allowed lazy loading of functions
			if isinstance(func, basestring):
				func = deferred_add(func)
				context[funname] = func
	elif hasattr(context, funname):
		func = getattr(context, funname)

	if func is None:
		raise NotFoundError("Function not found", function=funname)

	return func

@annotate.context("root")
class InitialContext(dict):
	pass

class HierarchicalShell:
	def __init__(self, name, no_rc=False):
		"""
		Create a new HierarchicalShell, optionally loading initialization
		statements from an RCFile based on the passed identifier 'name'.
		"""

		self.name = name
		self.init_commands = {}

		if not no_rc:
			self._load_rc()

		self.root = InitialContext()
		self.contexts = [self.root]

		self.root_add('import_types', import_types)

		#Initialize the root context if required
		self._check_initialize_context()

	def root_update(self, dict_like):
		"""
		Add all of the entries in the dictionary like object dict_like to the root
		context 
		"""

		self.root.update(dict_like)

	def root_add(self, name, value):
		"""
		Add a single function to the root context
		"""

		self.root[name] = value

	def context_name(self):
		"""
		Get the string name of the current context
		"""

		return annotate.context_name(self.contexts[-1])

	def finished(self):
		"""
		If there are no more contexts on the context list, then we cannot execute any
		commands and the shell has finished its useful life.
		"""

		return len(self.contexts) == 0

	def valid_identifiers(self):
		"""
		Return a list of the valid identifiers that can be called given the context that
		we are in.  Callers can use this result to perform autocomplete if they would like.
		"""

		funcs = annotate.find_all(self.contexts[-1]).keys() + builtin_help.keys()
		return funcs

	def _load_rc(self):
		"""
		Load context initialization commands from a configuration file.  The 
		file should have the format:
		[Context1]
		<list of commands>

		[OldContext1.Subcontext1]
		<list of commands>

		where list of commands is a list of lines that are executed using invoke
		whenever a context matching the identifier inside of the brackets is created
		for the first time.  This lets you run custom configuration routines to take 
		care of tiresome initialization calls automatically.
		"""

		rcfile = RCFile(self.name)
		context = None
		cmds = []

		for line in rcfile.contents:
			line = line.strip()

			if len(line) == 0:
				continue

			if line[0] == '[':
				lasti = line.find(']')
				if lasti == -1:
					raise InternalError("Syntax Error in rcfile, missing closing ]", line=line, name=self.name, path=rcfile.path)

				if context is not None:
					self.init_commands[context] = cmds
				
				#Start a new context
				context = line[1:lasti].strip()
				cmds = []
			else:
				#Process a command line
				if context is None:
					raise InternalError("Syntax Error in rcfile, command given before a context was specified", line=line, name=self.name, path=rcfile.path)

				cmds.append(line)

		if context is not None:
			self.init_commands[context] = cmds

	def _check_initialize_context(self):
		"""
		Check if our context matches something that we have initialization commands
		for.  If so, run them to initialize the context before proceeding with other
		commands.
		"""

		path = ".".join([annotate.context_name(x) for x in self.contexts])

		#Make sure we don't clutter up the output with return values from
		#initialization functions
		old_interactive = type_system.interactive
		type_system.interactive = False

		for key,cmds in self.init_commands.iteritems():
			if path.endswith(key):
				for cmd in cmds:
					line = shlex.split(cmd, posix=posix_lex)

					#Automatically remove enclosing double quotes on windows since they are not removed by shlex in nonposix mode
					def remove_quotes(x):
						if len(x) > 0 and x.startswith(("'", '"')) and x[0] == x[-1]:
							return x[1:-1]

						return x

					if not posix_lex:
						line = map(remove_quotes, line)

					self.invoke(line)

		type_system.interactive = old_interactive

	def invoke(self, line):
		"""
		Given a list of command line arguments, attempt to find the function being specified
		and map the passed arguments to that function based on its annotated type information. 
		"""

		funname = line[0]
		context = self.contexts[-1]

		#Check if we are asked for help
		if funname == 'help':
			return _do_help(context, line)
		if funname == 'quit':
			del self.contexts[:]
			return [], True
		if funname == 'back':
			self.contexts.pop()
			return [], True

		func = find_function(context, funname)

		#If this is a context derived from a module or package, just jump to it
		#since there is no initialization function
		if isinstance(func, annotate.BasicContext):
			self.contexts.append(func)
			self._check_initialize_context()
			return line[1:], False

		#find out how many position and kw args this function takes
		posset,kwset = annotate.get_spec(func)

		#If the function wants arguments directly, do not parse them
		if func.takes_cmdline == True:
			val = func(line[1:])
		else:
			arg_it = (x for x in line[1:])
			kwargs = {}
			posargs = []

			i = 1
			for arg in arg_it:
				if arg.startswith('--') or ((arg.startswith('-') and len(arg)==2)):
					name,val,skip = process_kwarg(arg, arg_it)
					kwargs[name] = val
					i+= skip
				else:
					if annotate.spec_filled(posset, kwset, posargs, kwargs):
						break

					posargs.append(arg)

				i += 1

			if not annotate.spec_filled(posset, kwset, posargs, kwargs):
				raise ArgumentError("too few arguments")

			val = func(*posargs, **kwargs)

		#Update our current context if this function destroyed it or returned a new one.
		finished = True

		if func.finalizer == True:
			self.contexts.pop()
		elif val is not None:
			if annotate.check_returns_data(func):
				if type_system.interactive:
					annotate.print_retval(func, val)
			else:
				self.contexts.append(val)
				self._check_initialize_context()
				finished = False

		return line[i:], finished



# This file is adapted from python code released by WellDone International
# under the terms of the LGPLv3.  WellDone International's contact information is
# info@welldone.org
# http://welldone.org
#
# Modifications to this file from the original created at WellDone International 
# are copyright Arch Systems Inc.

import os.path

def convert(arg):
	if arg is None:
		return None

	return str(arg)

def validate_readable(arg):
	if arg is None:
		raise ValueError("Path must be readable")

	if not os.path.isfile(arg):
		raise ValueError("Path is not a file")

	try:
		f = open(arg, "r")
		f.close()
	except:
		raise ValueError("Path could not be opened for reading")

def validate_exists(arg):
	if arg is None:
		raise ValueError("Path must exist")

	if not os.path.exists(arg):
		raise ValueError("Path must exist")

def validate_writeable(arg):
	if arg is None:
		raise ValueError("Path must be writable")

	parent = os.path.dirname(arg)
	if not os.path.isdir(parent):
		raise ValueError("Parent directory does not exist and path must be writeable")
	
#Formatting functions
def default_formatter(arg, **kwargs):
	return str(arg)


"""Virtual interfaces are ways to connect to a virtual IOTile device like a real one

Virtual things are different from Mock things in that mock things are designed specifically
to facilitate unit testing and hence do not allow for the complete configurability of behavior.
Virtual things are designed to allow the same level of configurability and robustness as a real
thing. 
"""

from Queue import Queue, Empty
import logging
from iotile.core.exceptions import ArgumentError
import audit

class VirtualIOTileInterface(object):
    """A virtual interface that presents an IOTile device to the world

    An example of a virtual interface is a bluetooth adapter configured with
    a GATT server that implements the TileBus over BLE protocol allowing any
    BLE client to connect to this virtual IOTile device as if it were a real
    one.

    Args:
        device (VirtualIOTileDevice): The actual device implementation that this
            virtual interface is providing access to.
    """

    def __init__(self):
        self.device = None
        self.audit_logger = logging.getLogger('virtual.audit')
        self.audit_logger.addHandler(logging.NullHandler())

        self.actions = Queue()
        self.reports = Queue()
        self.traces = Queue()

        self.in_progress_report = None
        self.in_progress_trace = None

    def start(self, device):
        """Begin allowing connections to a virtual IOTile device
        
        Args:
            device (VirtualIOTileDevice): The python object that implements the IOTile
                device functionality that we wish to allow interaction with.
        """

        raise NotImplementedError("VirtualIOTileInterface subclass did not override start")

    def process(self):
        """Process all pending actions, needs to be called periodically after start

        This function will block for up to 100 ms so that it may be simple called in a tight
        while loop without pegging the CPU to 100%.
        """

        try:
            while True:
                func, args = self.actions.get(timeout=0.1)
                func(*args)
        except Empty:
            pass

    def stop(self):
        """Stop allowing connections to this virtual IOTile device
        """
        
        raise NotImplementedError("VirtualIOTileInterface subclass did not override start")

    def _audit(self, event, **args):
        """Log an audit event with the given message and parameters

        Args:
            event (string): The event type that we are logging
            **args (dict): All required and any desired optional parameters associated
                with the audit event
        """

        if not hasattr(audit, event):
            raise ArgumentError("Unknown audit event name", name=event)

        audit_evt = getattr(audit, event)

        if len(args) > 0:
            self.audit_logger.info(audit_evt.message, args, extra={'event_name': audit_evt.name})
        else:
            self.audit_logger.info(audit_evt.message, extra={'event_name': audit_evt.name})

    def _defer(self, action, args=None):
        """Queue an action to be executed the next time process is called

        This is very useful for callbacks that should be called on the main thread but are queued
        from a different thread.

        Args:
            action (callable): A function to be called as action(*args)
            args (list): A list of arguments (possibly empty) to be passed to action
        """

        if args is None:
            args = []

        self.actions.put((action, args))

    def _queue_reports(self, *reports):
        """Queue reports for transmission over the streaming interface

        The primary reason for this function is to allow for a single implementation
        of encoding and chunking reports for streaming.

        Args:
            *reports (list): A list of IOTileReport objects that should be sent over
                the streaming interface.
        """

        for report in reports:
            self.reports.put(report)

    def _queue_traces(self, *traces):
        """Queue tracing information for transmission over the tracing interface

        The primary reason for this function is to allow for a single implementation
        of encoding and chunking traces for tracing

        Args:
            *traces (list): A list of IOTileReport objects that should be sent over
                the streaming interface.
        """

        for trace in traces:
            self.traces.put(trace)

    def _next_streaming_chunk(self, max_size):
        """Get the next chunk of data that should be streamed

        Args:
            max_size (int): The maximum size of the chunk to be returned

        Returns:
            bytearray: the chunk of raw data with size up to but not exceeding
                max_size.
        """

        chunk = bytearray()
        
        while len(chunk) < max_size:
            desired_size = max_size - len(chunk)

            #If we don't have an in progress report at the moment, attempt to get one
            if self.in_progress_report is None:
                try:
                    next_report = self.reports.get_nowait()
                except Empty:
                    return chunk

                self._audit('ReportStreamed', report=str(next_report))
                self.in_progress_report = bytearray(next_report.encode())

            if len(self.in_progress_report) <= desired_size:
                chunk += self.in_progress_report
                self.in_progress_report = None
            else:
                remaining = self.in_progress_report[desired_size:]
                chunk += self.in_progress_report[:desired_size]

                self.in_progress_report = remaining

        return chunk

    def _next_tracing_chunk(self, max_size):
        """Get the next chunk of data that should be streamed

        Args:
            max_size (int): The maximum size of the chunk to be returned

        Returns:
            bytearray: the chunk of raw data with size up to but not exceeding
                max_size.
        """

        chunk = bytearray()
        
        while len(chunk) < max_size:
            desired_size = max_size - len(chunk)

            #If we don't have an in progress report at the moment, attempt to get one
            if self.in_progress_trace is None:
                try:
                    next_trace = self.traces.get_nowait()
                except Empty:
                    return chunk

                self._audit('TraceSent', trace=str(next_trace))
                self.in_progress_trace = bytearray(next_trace)

            if len(self.in_progress_trace) <= desired_size:
                chunk += self.in_progress_trace
                self.in_progress_trace = None
            else:
                remaining = self.in_progress_trace[desired_size:]
                chunk += self.in_progress_trace[:desired_size]

                self.in_progress_trace = remaining

        return chunk



# This file is adapted from python code released by WellDone International
# under the terms of the LGPLv3.  WellDone International's contact information is
# info@welldone.org
# http://welldone.org
#
# Modifications to this file from the original created at WellDone International 
# are copyright Arch Systems Inc.

from iotile.core.hw.hwmanager import HardwareManager
from iotile.core.hw.reports.signed_list_format import SignedListReport
from iotile.core.hw.exceptions import *
from iotile.core.exceptions import *
import unittest
import pytest
import os.path
import os

@pytest.fixture
def simple_hw():
    hw = HardwareManager('virtual:simple')
    yield hw

    hw.disconnect()


@pytest.fixture
def report_hw():
    hw = HardwareManager('virtual:report_test')
    yield hw

    hw.disconnect()

@pytest.fixture
def conf_report_hw():
    conf_file = os.path.join(os.path.dirname(__file__), 'report_test_config_hash.json')

    if '@' in conf_file or ',' in conf_file or ';' in conf_file:
        pytest.skip('Cannot pass device config because path has [@,;] in it')

    hw = HardwareManager('virtual:report_test@%s' % conf_file)
    yield hw

    hw.disconnect()

@pytest.fixture
def conf2_report_hw():
    conf_file = os.path.join(os.path.dirname(__file__), 'report_test_config_signed.json')

    if '@' in conf_file or ',' in conf_file or ';' in conf_file:
        pytest.skip('Cannot pass device config because path has [@,;] in it')

    hw = HardwareManager('virtual:report_test@%s' % conf_file)
    yield hw

    hw.disconnect()

def test_basic(simple_hw):
    simple_hw.connect_direct('1')


def test_report(report_hw):
    report_hw.connect_direct('1')
    report_hw.enable_streaming()
    assert report_hw.count_reports() == 100

def test_config_file(conf_report_hw):
    """Make sure we can pass a config dict
    """

    conf_report_hw.connect_direct('2')
    conf_report_hw.enable_streaming()
    assert conf_report_hw.count_reports() == 11

def test_config_file(conf2_report_hw, monkeypatch):
    """Make sure we can pass a config dict
    """

    monkeypatch.setenv('USER_KEY_00000002', '0000000000000000000000000000000000000000000000000000000000000000')

    conf2_report_hw.connect_direct('2')
    conf2_report_hw.enable_streaming()

    assert conf2_report_hw.count_reports() == 11

    for report in conf2_report_hw.iter_reports():
        assert report.verified
        assert report.signature_flags == 1
        assert report.lowest_id >= 1
        assert report.highest_id > report.lowest_id
        assert isinstance(report, SignedListReport)



"""Auth providers are classes that can encrypt, decrypt, sign or verify data from IOTile devics

All auth providers must inherit from the AuthProvider class defined in this file and override the
methods that it provides with their own implementations.  
"""

import pkg_resources
from iotile.core.exceptions import NotFoundError, ArgumentError

HashOnlySHA256Signature = 0
HMACSHA256UserKey = 1
HMACSHA256FactoryKey = 2

KnownSignatureMethods = {
    HashOnlySHA256Signature:    'hash_only_sha256',        #Data is not signed, there is simply a sha256 integrity check
    HMACSHA256UserKey:          'hmac_sha256_user_key',    #Data is signed simply with HMAC using a user settable key on the device
    HMACSHA256FactoryKey:       'hmac_sha256_factory_key'  #Data is signed with HMAC using a factory set secret key associated with the device
}

class AuthProvider(object):
    """Base class for all objects that provide a way to authenticate or protect data
    
    Args:
        args (dict): An optional dictionary of AuthProvider specific config information
    """

    def __init__(self, args=None):
        if args is None:
            args = {}

        self.args = args

    @classmethod
    def FindByName(self, name):
        """
        """
        
        for entry in pkg_resources.iter_entry_points('iotile.auth_provider'):
            if entry.name == name:
                return entry.load()

    def encrypt(self, device_id, enc_method, data):
        """Encrypt a buffer of data on behalf of a device

        Args:
            device_id (int): The id of the device that we should encrypt for
            enc_method (int): The method of encryption that we should perform
            data (bytearray): The data that we should encrypt

        Returns:
            dict: The encrypted data and any associated metadata about the data.
                The data itself must always be a bytearray stored under the 'data'
                key, however additional keys may be present depending on the encryption method
                used.

        Raises:
            NotFoundError: If the auth provider is not able to encrypt the data.
        """

        raise NotFoundError("encrypt method is not implemented")

    def decrypt(self, device_id, enc_method, data):
        """Decrypt a buffer of data on behalf of a device

        Args:
            device_id (int): The id of the device that we should encrypt for
            enc_method (int): The method of encryption that we should perform
            data (bytearray): The data that we should encrypt

        Returns:
            dict: The decrypted data and any associated metadata about the data.
                The data itself must always be a bytearray stored under the 'data'
                key, however additional keys may be present depending on the encryption method
                used.

        Raises:
            NotFoundError: If the auth provider is not able to encrypt the data.
        """

        raise NotFoundError("decrypt method is not implemented")

    def sign(self, device_id, sig_method, data):
        """Sign a buffer of data on behalf of a device

        Args:
            device_id (int): The id of the device that we should encrypt for
            sig_method (int): The method of encryption that we should perform
            data (bytearray): The data that we should sign

        Returns:
            dict: The signature and any associated metadata about the signature.
                The signatured itself must always be a bytearray stored under the 
                'signature' key, however additional keys may be present depending 
                on the signature method used.

        Raises:
            NotFoundError: If the auth provider is not able to encrypt the data.
        """

        raise NotFoundError("sign method is not implemented")

    def verify(self, device_id, sig_method, data, signature):
        """Verify the signature attached to a buffer of data

        Args:
            device_id (int): The id of the device that we should encrypt for
            sig_method (int): The method of signing that was used
            data (bytearray): The data whose signature we should verify
            signature (bytearray): The signature attached to data

        Returns:
            dict: The result of the verification process must always be a bool under the
                'verified' key, however additional keys may be present depending on the 
                signature method used.

        Raises:
            NotFoundError: If the auth provider is not able to encrypt the data.
        """

        raise NotFoundError("verify method is not implemented")

    def _check_signature_method(self, sig_method):
        if sig_method in KnownSignatureMethods:
            return True

        raise NotFoundError("unknown signature method", method=sig_method)


from individual_format import IndividualReadingReport
from report import IOTileReading, IOTileReport
from signed_list_format import SignedListReport
from parser import IOTileReportParser

__all__ = ['IndividualReadingReport', 'IOTileReport', 'IOTileReading', 'SignedListReport', 'IOTileReportParser']



import argparse
import pkg_resources
import sys
import logging
import json
import imp
import os.path
import inspect
from iotile.core.hw.virtual.virtualdevice import VirtualIOTileDevice

def main():
    """Serve access to a virtual IOTile device using a virtual iotile interface
    """

    list_parser = argparse.ArgumentParser(add_help=False)
    list_parser.add_argument('-l', '--list', action='store_true', help="List all known installed interfaces and devices and then exit")

    parser = argparse.ArgumentParser(description="Serve acess to a virtual IOTile device using a virtual IOTile interface")
    
    parser.add_argument('interface', help="The name of the virtual device interface to use")
    parser.add_argument('device', help="The name of the virtual device to create")
    parser.add_argument('-c', '--config', help="An optional JSON config file with arguments for the interface and device")
    parser.add_argument('-l', '--list', action='store_true', help="List all known installed interfaces and devices and then exit")

    args, rest = list_parser.parse_known_args()

    if args.list:
        #List out known virtual interfaces
        print("Installed Virtual Interfaces:")
        for entry in pkg_resources.iter_entry_points('iotile.virtual_interface'):
            print('- {}'.format(entry.name))

        print("\nInstalled Virtual Devices:")
        for entry in pkg_resources.iter_entry_points('iotile.virtual_device'):
            print('- {}'.format(entry.name))

        return 0

    args = parser.parse_args()

    config = {}
    iface = None
    if args.config is not None:
        with open(args.config, "rb") as conf_file:
            config = json.load(conf_file)

    try:
        iface = instantiate_interface(args.interface, config)
        device = instantiate_device(args.device, config)

        handler = logging.StreamHandler()
        handler.setFormatter(logging.Formatter('%(asctime)s [AUDIT %(event_name)s] %(message)s'))

        iface.audit_logger.addHandler(handler)
        iface.audit_logger.setLevel(logging.INFO)

        iface.start(device)

        print("Starting to serve virtual IOTile device")

        #We need to periodically process events that are queued up in the interface
        while True:
            iface.process()

    except KeyboardInterrupt:
        print("Break received, cleanly exiting...")
    finally:
        if iface is not None:
            iface.stop()

def import_device_script(script_path):
    """Import a virtual device from a file rather than an installed module

    script_path must point to a python file ending in .py that contains exactly one
    VirtualIOTileDevice class definitions.  That class is loaded and executed as if it
    were installed.

    Args:
        script_path (string): The path to the script to load

    Returns:
        VirtualIOTileDevice: A subclass of VirtualIOTileDevice that was loaded from script_path
    """

    search_dir, filename = os.path.split(script_path)
    if search_dir == '':
        search_dir = './'

    if filename == '' or not os.path.exists(script_path):
        print("Could not find script to load virtual device, path was %s" % script_path)
        sys.exit(1)

    module_name, ext = os.path.splitext(filename)
    if ext != '.py':
        print("Script did not end with .py")
        sys.exit(1)

    try:
        file = None
        file, pathname, desc = imp.find_module(module_name, [search_dir])
        mod = imp.load_module(module_name, file, pathname, desc)
    finally:
        if file is not None:
            file.close()

    devs = filter(lambda x: inspect.isclass(x) and issubclass(x, VirtualIOTileDevice) and x != VirtualIOTileDevice, mod.__dict__.itervalues())
    if len(devs) == 0:
        print("No VirtualIOTileDevice subclasses were defined in script")
        sys.exit(1)
    elif len(devs) > 1:
        print("More than one VirtualIOTileDevice subclass was defined in script: %s" % str(devs))
        sys.exit(1)

    return devs[0]

def instantiate_device(virtual_dev, config):
    """Find a virtual device by name and instantiate it

    Args:
        virtual_dev (string): The name of the pkg_resources entry point corresponding to
            the device.  It should be in group iotile.virtual_device.  If virtual_dev ends
            in .py, it is interpreted as a python script and loaded directly from the script.
        config (dict): A dictionary with a 'device' key with the config info for configuring
            this virtual device.  This is optional.

    Returns:
        VirtualIOTileDevice: The instantiated subclass of VirtualIOTileDevice
    """

    conf = {}
    if 'device' in config:
        conf = config['device']

    #If we're given a path to a script, try to load and use that rather than search for an installed module 
    if virtual_dev.endswith('.py'):
        dev = import_device_script(virtual_dev)
        return dev(conf)

    for entry in pkg_resources.iter_entry_points('iotile.virtual_device', name=virtual_dev):
        dev = entry.load()
        return dev(conf)

    print("Could not find an installed virtual device with the given name: {}".format(virtual_dev))
    sys.exit(1)

def instantiate_interface(virtual_iface, config):
    """Find a virtual interface by name and instantiate it

    Args:
        virtual_iface (string): The name of the pkg_resources entry point corresponding to
            the device.  It should be in group iotile.virtual_device
        config (dict): A dictionary with a 'device' key with the config info for configuring
            this virtual device.  This is optional.

    Returns:
        VirtualInterface: The instantiated subclass of VirtualInterface
    """

    conf = {}
    if 'device' in config:
        conf = config['device']

    for entry in pkg_resources.iter_entry_points('iotile.virtual_interface', name=virtual_iface):
        dev = entry.load()
        return dev(conf)

    print("Could not find an installed virtual interface with the given name: {}".format(virtual_iface))
    sys.exit(1)



# This file is adapted from python code released by WellDone International
# under the terms of the LGPLv3.  WellDone International's contact information is
# info@welldone.org
# http://welldone.org
#
# Modifications to this file from the original created at WellDone International 
# are copyright Arch Systems Inc.
import pkg_resources

from iotile.core.utilities.typedargs import *
from Queue import Empty
from iotile.core.hw.transport import *
from iotile.core.hw.exceptions import *
from iotile.core.exceptions import *
from iotile.core.utilities.typedargs.annotate import param, return_type, finalizer
from iotile.core.hw.proxy.proxy import TileBusProxyObject
from iotile.core.dev.registry import ComponentRegistry
from iotile.core.hw.transport.adapterstream import AdapterCMDStream
import re
import inspect
import os.path
import imp
import binascii
import sys
from iotile.core.utilities.packed import unpack

@context("HardwareManager")
class HardwareManager:
    """
    A module for managing and interacting with IOTile Hardware

    This context provides tools to configure, control, debug and program
    any IOTile module.  Specific functionality can be implemented in dynamically
    loaded proxy objects that are designed to provide access to each IOTile.

    To create a HardwareManager, you need to pass a port string that describes the 
    method to be used to connect to the IOTile device. The method should specify the 
    name of the connection method optionally followed by a colon and any extra information
    possibly needed to connect using that method.
    """

    @param("port", "string", desc="transport method to use in the format transport[:port[,connection_string]]")
    @param("record", "path", desc="Optional file to record all RPC calls and responses made on this HardwareManager")
    def __init__(self, port="none", record=None):
        transport, _, arg = port.partition(':')

        self.transport = transport
        self.port = None
        if arg != "":
            self.port = arg

        self.record = record
        
        self.stream = self._create_stream()
        self._stream_queue = None
        self._trace_queue = None
        self._trace_data = bytearray()
        self.proxies = {'TileBusProxyObject': TileBusProxyObject}
        self.name_map = {TileBusProxyObject.ModuleName(): [TileBusProxyObject]}

        self._setup_proxies()

    def _setup_proxies(self):
        """
        Load in proxy module objects for all of the registered components on this system
        """

        # Find all of the registered IOTile components and see if we need to add any proxies for them
        reg = ComponentRegistry()
        modules = reg.list_components()

        proxies = reduce(lambda x,y: x+y, [reg.find_component(x).proxy_modules() for x in modules], [])
        for prox in proxies:
            self.add_proxies(prox)

        # Find all installed proxy objects through registered entry points
        for entry in pkg_resources.iter_entry_points('iotile.proxy'):
            mod = entry.load()
            self._add_proxy_module(mod)

    @param("address", "integer", "positive", desc="numerical address of module to get")
    def get(self, address):
        """
        Create a proxy object for a tile by address.

        The correct proxy object is determined by asking the tile for its status information
        and looking up the appropriate proxy in our list of installed proxy objects
        """

        tile = self._create_proxy('TileBusProxyObject', address)
        name = tile.tile_name()
        version = tile.tile_version()

        # Now create the appropriate proxy object based on the name and version of the tile 
        tile_type = self.get_proxy(name)
        if tile_type is None:
            raise HardwareError("Could not find proxy object for tile", name="'{}'".format(name), known_names=self.name_map.keys())

        tile = tile_type(self.stream, address)      
        return tile

    @annotated
    def controller(self):
        """
        Find an attached IOTile controller and attempt to connect to it.
        """

        con = self.get(8)       
        con._hwmanager = self

        return con

    @param("device_uuid", "integer", desc="UUID of the device we would like to connect to")
    def connect(self, device_uuid):
        """Attempt to connect to a device by its UUID
        """

        self.stream.connect(device_uuid)

    @param("connection_string", "string", desc="opaque connection string indicating which device")
    def connect_direct(self, connection_string):
        """Attempt to connect to a device using a connection string
        """

        self.stream.connect_direct(connection_string)

    @annotated
    def disconnect(self):
        """Attempt to disconnect from a device 
        """

        self.stream.disconnect()

    @return_type("bool")
    def heartbeat(self):
        """
        Check whether the underlying command stream is functional
        """

        return self.stream.heartbeat()

    @annotated
    def enable_streaming(self):
        """
        Enable streaming of sensor graph data over this interface
        """

        self._stream_queue = self.stream.enable_streaming()

    @annotated
    def enable_tracing(self):
        """Enable tracing of realtime debug information over this interface
        """

        self._trace_queue = self.stream.enable_tracing()

    @return_type("integer")
    def count_reports(self):
        if self._stream_queue is None:
            return 0

        return self._stream_queue.qsize()

    @return_type("string")
    @param("encoding", "string", desc="The encoding to use to dump the trace, either 'hex' or 'raw'")
    def dump_trace(self, encoding):
        """Dump all received tracing data currently received from the device to stdout

        The data is encoded per the encoding parmeter which must be either
        the string 'hex' or 'raw'.  If hex is passed, the data is printed as hex digits,
        if raw is passed, the data is printed as received from the device.
        """

        if encoding not in ['raw', 'hex']:
            raise ValidationError("Unknown encoding type specified in dump trace", encoding=encoding, known_encodings=['hex', 'raw'])

        if self._trace_queue is None:
            return ""

        self._accumulate_trace()

        if encoding == 'raw':
            return str(self._trace_data)

        return binascii.hexlify(self._trace_data)

    def _accumulate_trace(self):
        """Copy tracing data from trace queue into _trace_data
        """

        if self._trace_queue is None:
            return

        try:
            while True:
                blob = self._trace_queue.get(block=False)
                self._trace_data += bytearray(blob)
        except Empty:
            pass

    def iter_reports(self, blocking=False):
        """Iterate over reports that have been received

        If blocking is True, this iterator will never stop.  Otherwise
        it will iterate over all reports currently in the queue (and those
        added during iteration)

        Args:
            blocking (bool): Whether to stop when there are no more readings or
                block and wait for more.
        """
        if self._stream_queue is None:
            return

        try:
            while True:
                yield self._stream_queue.get(block=blocking)
        except Empty:
            pass

    def wait_reports(self, num_reports, timeout=2.0):
        """Wait for a fixed number of reports to be received

        Args:
            num_reports (int): The number of reports to wait for
            timeout (float): The maximum number of seconds to wait without
                receiving another report.
        """

        if self._stream_queue is None:
            raise EnvironmentError("You have to enable streaming before you can wait for reports")

        reports = []

        for i in xrange(0, num_reports):
            try:
                report = self._stream_queue.get(timeout=timeout)
                reports.append(report)
            except Empty:
                raise TimeoutError("Timeout waiting for a report", expected_number=num_reports, received_number=i, received_reports=reports)

        return reports

    @annotated
    def reset(self):
        """
        Attempt to reset the underlying stream back to a known state
        """

        self.stream.reset()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.stream.close()
        return False

    @finalizer
    def close(self):
        self.stream.close()

    @param("path", "path", "readable", desc="path to module to load")
    @return_type("integer")
    def add_proxies(self, path):
        """
        Add all proxy objects defined in the python module located at path.

        The module is loaded and all classes that inherit from TileBusProxyObject
        are loaded and can be used to interact later with modules of that type.

        Returns the total number of proxies added.
        """

        d,p = os.path.split(path)

        p,ext = os.path.splitext(p)
        if ext != '.py' and ext != '.pyc' and ext != "":
            raise ArgumentError("Passed proxy module is not a python package or module (.py or .pyc)", path=path)

        try:
            fileobj,pathname,description = imp.find_module(p, [d])

            #Don't load modules twice
            if p in sys.modules:
                mod = sys.modules[p]
            else:
                mod = imp.load_module(p, fileobj, pathname, description)
        except ImportError as e:
            raise ArgumentError("could not import module in order to load external proxy modules", module_path=path, parent_directory=d, module_name=p, error=str(e))

        return self._add_proxy_module(mod)

    def _add_proxy_module(self, mod):
        """Add a proxy module that has already been imported

        Args:
            mod (module): A python module object that may contain a TileBusProxyObject subclass

        Returns:
            integer: The number of new TilebusProxyObject classes found
        """

        num_added = 0
        for obj in filter(lambda x: inspect.isclass(x) and issubclass(x, TileBusProxyObject) and x != TileBusProxyObject, mod.__dict__.itervalues()):
            if obj.__name__ in self.proxies:
                continue #Don't readd proxies that we already know about

            self.proxies[obj.__name__] = obj

            #Check if this object matches a specific shortened name so that we can 
            #automatically match a hw module to a proxy without user intervention
            if hasattr(obj, 'ModuleName'):
                short_name = obj.ModuleName()
                if short_name in self.name_map:
                    self.name_map[short_name].append(obj)
                else:
                    self.name_map[short_name] = [obj]

            num_added += 1

        return num_added

    @annotated
    @return_type("list(basic_dict)")
    def scan(self):
        """
        Scan for available devices and print a dictionary of information about them
        """

        devices = self.stream.scan()

        #FIXME: Use dictionary format in bled112stream to document information returned about devices
        return devices

    def get_proxy(self, short_name):
        """
        Find a proxy type given its short name.

        If no proxy type is found, return None.
        """

        if short_name not in self.name_map:
            return None

        return self.name_map[short_name][0]

    def _create_proxy(self, proxy, address):
        """
        Create a python proxy object to talk to an IOTile module with the given type
        at the given address.
        """

        if proxy not in self.proxies:
            raise UnknownModuleTypeError("unknown proxy module specified", module_type=proxy, known_types=self.proxies.keys())

        proxy_class = self.proxies[proxy]
        return proxy_class(self.stream, address)

    def _create_stream(self):
        conn_string = None
        port = self.port
        if port is not None and "," in port:
            port, conn_string = port.split(',')

        if port is not None:
            port = port.strip()
        if conn_string is not None:
            conn_string = conn_string.strip()

        #First check if this is the special none stream that creates a transport channel nowhere
        if self.transport == 'none':
            return CMDStream(port, conn_string, record=self.record)

        #Next attempt to find a CMDStream that is registered for this transport type
        for stream_entry in pkg_resources.iter_entry_points('iotile.cmdstream'):
            if stream_entry.name != self.transport:
                continue

            stream_factory = stream_entry.load()
            return stream_factory(port, conn_string, record=self.record)

        #Otherwise attempt to find a DeviceAdapter that we can turn into a CMDStream
        for adapter_entry in pkg_resources.iter_entry_points('iotile.device_adapter'):
            if adapter_entry.name != self.transport:
                continue

            adapter_factory = adapter_entry.load()
            return AdapterCMDStream(adapter_factory(port), port, conn_string, record=self.record)

        raise HardwareError("Could not find transport object registered to handle passed transport type", transport=self.transport)
